/*
 * Copyright (C) 2009-2010 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _AMISH_H
#define _AMISH_H

#include <avr/pgmspace.h>
#include <inttypes.h>
#include <ctype.h>
#include "config.h"
#include "avrutils.h"

/*
 * AMISH
 * AVR Multi-Interface SHell
 * Process input commands and return output. AMISH may be used with
 * TCP/UDP and a serial console.
 */

/*
 * Define AMISH_SHORT_IO to reduce the maximum I/O size to 126 bytes.
 * Without this definition, 2^16-2 byte transfers are legal, theoretically.
 */

#define AMISH_SHORT_IO

#ifdef AMISH_SHORT_IO
	typedef uint8_t amish_io_len;
	#define AMISH_RET_STATE_FAILURE 127
#else
	typedef uint16_t amish_io_len;
	#define AMISH_RET_STATE_FAILURE 32767
#endif

/* Saved state for multiple calles of one command: */
typedef uint32_t8 amish_cmd_state_t;

/*
 * amish_ret:
 * Return value of amish_process / amish cmd handler.
 * The function may indicate with call_again that it wants to be called
 * again later. If the state equals AMISH_RET_STATE_FAILURE, the command
 * failed. Len is the written length of the output in bytes (including \0).
 */
typedef union {
	struct {
#ifdef AMISH_SHORT_IO
		amish_io_len len:7;
#else
		amish_io_len len:15;
#endif
		uint8_t call_again:1;
	};
	amish_io_len state;
	amish_io_len val;
} amish_cmd_ret;

typedef struct {
	int8_t arg;
	char* input;
	char* output;
	amish_cmd_state_t state;
	amish_io_len len;
	amish_cmd_ret ret;
} amish_cmd_arg;

/*
 * AMISH command entry:
 * For every command, there exists an command entry that specifies
 * its name and the function to be called for this name.
 * An additional argument may be used to multiplex multiple commands
 * with a single function.
 */
typedef struct {
	char* name;
	int8_t arg;
#define AMISH_HANDLER_ARGS (amish_cmd_arg*)
#define AMISH_HANDLER_FPOINTER(handler) void (*handler) AMISH_HANDLER_ARGS
	AMISH_HANDLER_FPOINTER(handler);
} amish_cmd;


typedef uint32_t8 amish_cmd_state;
#define AMISH_CMD_STATE(in) ((amish_cmd_state*)(in))

extern const amish_cmd amish_cmds[] PROGMEM;

/*
 * amish_init:
 * Initialize the internal amish structures.
 */
void amish_init(void);

/*
 * amish_worker:
 * Do periodic work as command processing.
 */
void amish_worker(void);

/*
 * amish_firstexec_cmd:
 * Process the input string "input" and return the associated function and
 * its first result.
 *
 * Beware that the input string may be abused (overwritten) by a command 
 * to save a temporary state between calls (amish_cmd_state)!
 */
amish_cmd* amish_firstexec_cmd(amish_cmd_arg* cmdarg);

/*
 * amish_reexec_cmd:
 * Repeat the execution of a command to obtain more output.
 * 
 */
void amish_reexec_cmd(amish_cmd* cmd, amish_cmd_arg *cmdarg);

/* AMISH-supplied commands: */
void amish_cmd_help(amish_cmd_arg*);

#define CMD_SYS_RESET 0
#define CMD_SYS_INFO 1
#define CMD_SYS_PW 2
#define CMD_SYS_EXIT 3
#define CMD_SYS_BL 4
void amish_cmd_sys(amish_cmd_arg*);


#define skip_whitespaces(str) \
	while(isspace(*str)){++str;}

/*
 * amish_parse_boolean:
 * returns 0 or 1, depending on the string or -1 if it was not recognised.
 */
int8_t amish_parse_boolean(const char* arg);

#endif
