/*
 * Copyright (C) 2009-2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "avrutils.h"
#include "stdio.h"
#include "amish.h"
#include "amish_uart.h"
#include <string.h>

/*
 * Reserve two extra bytes for sending \n\r.
 */
static char rxbuf[AMISH_UART_RXBUF];
static char txbuf[AMISH_UART_TXBUF + 3];
static char txringbuf[AMISH_UART_RINGBUF];
static uint8_t rxlen;
static uint8_t ring_head, ring_tail;
static amish_cmd* current_cmd;
static amish_cmd_state_t cmd_state;
static struct {
	uint8_t run_cmd:1;
	uint8_t non_first_call:1;
} amish_uart_info;

/* Mask used to calculate the modulo addition: */
#define UART_RINGBUF_MASK (AMISH_UART_RINGBUF - 1)
#define UART_RINGADD(start, len) (((start)+(len)) & UART_RINGBUF_MASK)

void uart_send(uint8_t data) {
	uint8_t new_head = UART_RINGADD(ring_head, 1);

	/* Wait for free space in the ring buffer: */
	while(vu8(ring_tail) == new_head);

	/* Store the data, then advance the buffer. */
	txringbuf[new_head] = data;
	vu8(ring_head) = new_head;

		/* The transmission may have stalled. Restart it: */
	UCSR0B |= (1 << UDRIE0);
}

void uart_send_string(char* str) {
	while(*str) {
		uart_send(*str);
		str++;
	}
}

int uart_putchar( char c, FILE *stream ) {
	if( c == '\n' )
			uart_putchar( '\r', stream );
#ifdef DEBUGGING_UART
	loop_until_bit_is_set( UCSR0A, UDRE0 );
	UDR0 = c;
#else
	uart_send(c);
#endif

	return 0;
}

static FILE ttyout = FDEV_SETUP_STREAM( uart_putchar, NULL, _FDEV_SETUP_WRITE );

/*
 * UART transmit completion interrupt service routine:
 * Advance the write pointer and push out the next byte.
 */
ISR(USART0_UDRE_vect) {

	if(ring_tail != ring_head) { /* There is data to transmit */
		uint8_t new_tail = UART_RINGADD(ring_tail, 1);
		ring_tail = new_tail;

		UDR0 = txringbuf[new_tail]; /* Now, push out the byte. */
	} else { /* All data transmitted */
		UCSR0B &= ~(1 << UDRIE0); /* Disable the data register empty interrupt. */
	}
}

ISR(USART0_RX_vect) {
	uint8_t recvbyte = UDR0;

	/* Suppress error conditions: */
	if((UCSR0A & ((1 << FE0) | (1 << DOR0) | (1 << UPE0)))) {
		/* Clear the UART data register by reading it: */
		(void) recvbyte;

	/* Else ensure that no command is currently being sent. */
	} else if(!amish_uart_info.run_cmd) {
		/*
		 * Trigger on any line ending, as a sole line ending is filtered by
		 * the worker logic.
		 */
		if(recvbyte == '\r' || recvbyte == '\n' || rxlen == AMISH_UART_RXBUF) {
			/* The command completion condition is met. */

			/* Send out a full line feed: */
			loop_until_bit_is_set(UCSR0A, UDRE0); /* XXX: Such a delay is ugly. */
			UDR0 = '\r';
			loop_until_bit_is_set(UCSR0A, UDRE0); /* XXX: Such a delay is ugly. */
			UDR0 = '\n';
			loop_until_bit_is_set(UCSR0A, UDRE0); /* XXX: Such a delay is ugly. */

			rxbuf[rxlen] = '\0';
			amish_uart_info.non_first_call = 0;
			amish_uart_info.run_cmd = 1;
		} else if ((recvbyte == '\b' || recvbyte == 127) && rxlen > 0) {
			/* Backspace can be either \b or DEL. Depends on your terminal. */
			/* A backspace can be handled, the buffer is not empty. */
			rxlen--;
			loop_until_bit_is_set(UCSR0A, UDRE0); /* XXX: Such a delay is ugly. */
			UDR0 = '\b'; /* Echo \b to erase the character on the terminal. */
			loop_until_bit_is_set(UCSR0A, UDRE0); /* XXX: Such a delay is ugly. */
			UDR0 = ' '; /* Echo \b to erase the character on the terminal. */
			loop_until_bit_is_set(UCSR0A, UDRE0); /* XXX: Such a delay is ugly. */
			UDR0 = '\b'; /* Echo \b to erase the character on the terminal. */
			loop_until_bit_is_set(UCSR0A, UDRE0); /* XXX: Such a delay is ugly. */
		} else if(recvbyte >= 32 && rxlen < AMISH_UART_RXBUF ) {
			/* Store the received character: */
			rxbuf[rxlen++] = recvbyte;
			/* And push the received character back as console echo: */
			loop_until_bit_is_set(UCSR0A, UDRE0); /* XXX: Such a delay is ugly. */
			UDR0 = recvbyte;
		}
	}
}

void amish_uart_init(void) {
	rxlen = 0;
	ring_head	= 0;
	ring_tail = 0;
	bzero(&amish_uart_info, sizeof(amish_uart_info));
#define UART_UBRR_CALC(BAUD_,FREQ_) ((FREQ_)/((BAUD_)*16L)-1)
	/*
	 * Initialize /all/ UART registers for this port. So it works even
	 * after the bootloader exits.
	 */
	UCSR0A = _BV(TXC0);
	UCSR0B = (1<<TXEN0) | (1<<RXEN0) | (1 << RXCIE0) ;		// UART TX, RX enabled
	UCSR0C = (1<<USBS0)|(3<<UCSZ00);		// 8N1
	UBRR0H = (uint8_t)( UART_UBRR_CALC( UART_BAUD_RATE, F_CPU ) >> 8 );
	UBRR0L = (uint8_t)UART_UBRR_CALC( UART_BAUD_RATE, F_CPU );
	stdout = &ttyout;
}


void amish_uart_worker(void) {
	/* Is there anything to do and the TX queue empty? */
	if(amish_uart_info.run_cmd) {

		/* Noise / empty command suppression: */
		if(rxlen < 2) {
			if(!amish_uart_info.non_first_call) goto print_prompt;
		}

		amish_cmd_arg ca;
		ca.input = rxbuf;
		ca.output = txbuf;
		ca.len = AMISH_UART_TXBUF;
		if(!amish_uart_info.non_first_call) {
			/* Call the requested command */
			current_cmd = amish_firstexec_cmd(&ca);
			amish_uart_info.non_first_call = 1;
		} else {
			ca.state = cmd_state;
			amish_reexec_cmd(current_cmd, &ca);
		}
		cmd_state = ca.state;

		/* Add a linebreak: */
		txbuf[ca.ret.len] = '\r';
		txbuf[++ca.ret.len] = '\n';
		txbuf[++ca.ret.len] = '\0';

		/* Put the string into the send queue: */
		uart_send_string(txbuf);

		/* Last call of the current command? */
		if(!ca.ret.call_again) {
print_prompt:
			rxlen = 0;
			amish_uart_info.non_first_call = 0;
			amish_uart_info.run_cmd = 0;
			strcpy_P(txbuf, PSTR(AMISH_PROMPT "> "));
			uart_send_string(txbuf);
		}
	}
}

