/*
 * Copyright (C) 2009-2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <avr/wdt.h>
#include <stdlib.h>
#include <limits.h>
#include "avrutils.h"

/*
 * crc8:
 * Generate an CRC8 value.
 */
uint8_t crc8 (uint8_t crc, uint8_t data) {
	uint8_t val = data ^ crc;
	crc = 0;
	if( val & 0x01 ) crc	= 0x5E; /* 0 ^ x = x */
	if( val & 0x02 ) crc ^= 0xBC;
	if( val & 0x04 ) crc ^= 0x61;
	if( val & 0x08 ) crc ^= 0xC2;
	if( val & 0x10 ) crc ^= 0x9D;
	if( val & 0x20 ) crc ^= 0x23;
	if( val & 0x40 ) crc ^= 0x46;
	if( val & 0x80 ) crc ^= 0x8C;
	return crc;
}

uint8_t gen_crc8(uint8_t *data, uint16_t len) {
	uint8_t crc = 0;
	uint16_t i = 0;
	do {
		crc = crc8(crc, *(data + i));
	} while (++i < len);
	return crc;
}

void noinline clear32(uint32_t *num) {
	*num = 0;
}

void noinline inc32(uint32_t *accumulator) {
	*accumulator += 1;
}

void noinline add32_16(uint32_t *accumulator, uint16_t value) {
	*accumulator += value;
}

void noinline sub32_16(uint32_t *accumulator, uint16_t value) {
	*accumulator -= value;
}

void noinline add16_16(uint16_t *accumulator, uint16_t value) {
	*accumulator += value;
}

void noinline sub16_16(uint16_t *accumulator, uint16_t value) {
	*accumulator -= value;
}

void reboot_device(void) {
	/* Disable interrupts, enable the watchdog and wait for it to fire. */
	cli();
	wdt_enable(WDTO_15MS);
	for(;;);
}

void reboot_to_bootloader(void) {
	void (*bl)(void) = (void*) ((uint32_t)FLASHEND+1) - 2 * REBOOT_BL_SIZE;
	/*
	 * XXX: For this to work, the internal pullup must not be used by the
	 * bootloader, but it has to be an external resistor!
	 */
	cli();
	SET_OUTPUT(REBOOT_BL_BUTTON);
	if(REBOOT_BL_ACTIVE)
		SET(REBOOT_BL_BUTTON);
	else
		RESET(REBOOT_BL_BUTTON);
	bl();
}

const char* parse_num32(const char* str, uint32_t* dst, uint8_t base) {
	uint32_t val = 0;
	const char* start = str;
	if(!str) return NULL;
	for(;;) {
		uint8_t n = *str;
		if(n < '0') break;
		if(n >= '0' && n <= '9') n -= '0';
		else if(n >= 'A' && n <= 'Z') n -= 'A' - 10;
		else n -= 'a' - 10;
		if(n >= base) break;
		val *= base;
		val += n;
		str++;
	}
	if(str == start) return NULL; /* No digit could be parsed */
	*dst = val;
	return str;
}

struct parse_ret16 parse_num16(const char* start, uint8_t base) {
	struct parse_ret16 ret;
	uint32_t val;
	ret.next = (char*)parse_num32(start, &val, base);
	ret.val = (uint16_t) val;
	return ret;
}

char* print_num32(char* start, uint32_t val, uint8_t base) {
	uint8_t i = 1;
	uint32_t val_ = val / base;

	/* Calculate largest digit: */
	while(val_) {
		val_ /= base;
		i++;
	}
	start += i;
	*start = '\0';
	char* end = start;

	/* Output the digits from left to right: */
	do {
		uint8_t n = val % base;
		val /= base;
		/* Adjust the digit n to an ASCII value: */
		if(n <= 9) n += '0';
		else if(n <= 'Z') n += 'A' - 10;
		else n += 'a' - 26 - 10;
		*--start = n;
	} while(--i);
	return end;
}

char* print_num16(char* start, uint16_t val, uint8_t base) {
	/* XXX: maybe fill in to make it faster? */
	return print_num32(start, val, base);
}

char* stpcpy_P(char* dst, const char* src_P) {
	src_P--;
	dst--;
	while((*++dst = pgm_read_byte(++src_P)));
	return dst;
}

char* stpcpy(char* dst, const char* src) {
	src--;
	dst--;
	while((*++dst = *(++src)));
	return dst;
}
