/*
 * Copyright (C) 2009-2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* enc0 is the prefix for enc28j60, enc1 will be enc[46]24j600. Later. ;-) */
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "config.h"
#define ENC0_INTERNAL
#include "enc0.h"
#include "avrutils.h"
#include "spi.h"
#include "uip/uip_arp.h"
#include "uip/uip-split.h"
#include "timer.h"

/* tunables:
 * - SPI interface (config.h):
 *	 - SPI_PRESCALER
 *	 - (ENC0_RESET), ENC0_SCK, ENC0_MISO, ENC0_MOSI, ENC0_CS
 * - duplex operation: ENC0_FULL_DUPLEX (beware! no autoconfig, untested)
 */
#ifndef ENC0_FULL_DUPLEX
	#define ENC0_FULL_DUPLEX 0
#endif

/* SPI configuration: sample on rising edge and CPHA=0 */

struct enc0_info {
	uint8_t int_pending:1;
	uint8_t tx_active:1;
	uint8_t tx_tries:2;
	uint8_t bank: 2;
};

/* Revision of the attached controller: */
static	uint8_t enc0_hw_rev;
/* Location of the next received packet: */
static uint16_t next_packet_loc;

struct enc0_info enc0_info;


timer_millistamp enc0_last_tx_stamp;

/*
 * Shortcut to select the controller:
 * Remember that the ENC28J60 needs at least 250 ns after the last bit is sent over
 * the SPI bus before CS may be cleared, if a PHY register was accessed.
 */
#define CS_ON	RESET(ENC0_CS)
#if F_CPU > 8000000
# define CS_OFF	 do { _delay_us(1); SET(ENC0_CS); } while(0)
#else
# define CS_OFF	 SET(ENC0_CS)
#endif

/*
 * Cheat a bit with the ENC0_OP_BFS/BFC. They only differ in a bit, so
 * make them a macro.
 */
#define enc0_bfs_reg(address, mask) \
	enc0_wbfx_reg(ENC0_OP_BFS, address, mask)
#define enc0_bfc_reg(address, mask) \
	enc0_wbfx_reg(ENC0_OP_BFC, address, mask)

/* internal forward declarations: */
static void		 enc0_soft_reset(void);
static uint8_t	enc0_read_reg(uint8_t address);
static void		 enc0_write_reg(uint8_t address, uint8_t value);
static void		 enc0_wbfx_reg(uint8_t opcode, uint8_t address, uint8_t value);
static void		 enc0_select_bank(uint8_t address);
static uint16_t enc0_read_regpair(uint8_t address);
static void		 enc0_write_regpair(uint8_t address, uint16_t value);
static void		 enc0_write_phy(uint8_t address, uint16_t value);
static uint16_t enc0_read_phy(uint8_t address);
uint8_t				 enc0_read_mem_byte(void);
void						enc0_read_mem_block(void* buffer, uint16_t length);
void						enc0_write_mem_byte(uint8_t data);
void						enc0_write_mem_block(void* buffer, uint16_t length);
static void		 enc0_recv_packet(void);
void						enc0_reset_tx(void);
void						enc0_retransmit(void);



void enc0_init(void) {
	/* Set up the I/O pins */

	/* Enable the controller by clearing the reset pin, wait 1 ms (Erratum 1): */
#ifdef ENC0_RESET
	RESET(ENC0_RESET);
	SET_OUTPUT(ENC0_RESET);
	_delay_ms(1);
	SET(ENC0_RESET);
#else
	_delay_ms(1);
#endif

	/*
	 * Disable the pullup on the input and bring outputs in a
	 * defined state for SPI
	 */
	RESET(ENC0_MISO);
	RESET(ENC0_MOSI);
	RESET(ENC0_SCK);
	SET(ENC0_CS);
	SET_INPUT(ENC0_MISO);
	SET_OUTPUT(ENC0_MOSI);
	SET_OUTPUT(ENC0_SCK);
	SET_OUTPUT(ENC0_CS);

	spi_init();

	/* Do a soft reset after the hard reset above, just in case. */
	enc0_soft_reset();

	/* We are at bank 0 after a reset. */
	enc0_info.bank = 0;

	/* Remember the hardware revision: */
	enc0_hw_rev = enc0_read_reg(EREVID);
	char buf[26]; char* b;
	b = stpcpy_P(buf, PSTR("ENC0: ENC28J60, rev. "));
	print_num16(b, enc0_hw_rev, 10);
	puts(buf);

	/* ETH buffer configuration: */

	/* Read buffer spans from 0 to 4 kB - 1: */
	enc0_write_regpair(ERXST, ENC0_RXBUF_START);
	enc0_write_regpair(ERXND, ENC0_RXBUF_END);

	/* Read pointer is set to the beginning. */
	next_packet_loc = ENC0_RXBUF_START;
	enc0_write_regpair(ERXRDPT, next_packet_loc);
	/* RX hardware write pointer already initialized to 0 by reset. */

	/* Transmit buffer spans from 4 kB to 8 kB: */
	enc0_write_regpair(ETXST, ENC0_TXBUF_START);
	/* Tx buffer end already initialized to 8 kB by reset. */


	/* Configure the MAC part: */

	/* Enable the receiver */
#define EXPECTED_MACON1 (MACON1_TXPAUS | MACON1_RXPAUS | MACON1_MARXEN)
	enc0_write_reg(MACON1, EXPECTED_MACON1);

	/*
	 * Pad packets for standard compatibility, enable CRC generation
	 * and frame length checking.
	 * The datasheet gives us these magic values for the inter-packet gap.
	 */
#if !ENC0_FULL_DUPLEX
# define EXPECTED_MACON3 (MACON3_PADCFG0 | MACON3_TXCRCEN | MACON3_FRMLNEN)
	enc0_write_reg(MACON3, EXPECTED_MACON3);
	enc0_write_reg(MABBIPG, 0x12);
#else
# define EXPECTED_MACON3 \
		(MACON3_PADCFG0 | MACON3_TXCRCEN | MACON3_FRMLNEN | MACON3_FULDPX)
	enc0_write_reg(MACON3, EXPECTED_MACON3);
	enc0_write_reg(MABBIPG, 0x15);
#endif

	/* now for the rest of the gap registers (see datasheet): */
	enc0_write_regpair(MAIPG, 0x0C12);

	/*
	 * Increase the value at which late collisions occur because of a
	 * silicon bug in rev. B5
	 */
	enc0_write_reg(MACLCON2, 63);

	/* Maximum packet length: UIP_MAX_ETH_FRAME_LENGTH + CRC */
	enc0_write_regpair(MAMXFL, UIP_MAX_ETH_FRAME_LENGTH + 4);

	/* Load the MAC address */
	enc0_update_mac();

	/*
	 * By default, we get unicast and broadcast packets with correct CRC.
	 * So there is no need to modify any filters.
	 */

	/* Disable the clock output. It only produces EMI if not needed. */
	enc0_write_reg(ECOCON, 0);

	/* Disable the hardware loopback. */
	enc0_write_phy(PHCON2, PHCON2_HDLDIS);

	/* LEDB: Link status, LEDA: Traffic */
	enc0_write_phy(PHLCON, PHLCON_LBCFG2 | PHLCON_LACFG2 | PHLCON_LACFG1 |
		PHLCON_LACFG0 | PHLCON_STRCH);

	/* Enable auto-increment to ease our life. */
	enc0_bfs_reg(ECON2, ECON2_AUTOINC);

	/*
	 * Enable interrupt generation for almost everything
	 * (TX, RX, Link change, errors).
	 * Enable the interrupt flags to get notifications about events, even
	 * if no IRQ pin is used.
	 */
	enc0_write_reg(EIE, EIE_PKTIE | EIE_LINKIE | EIE_TXIE | EIE_TXERIE |
#ifdef ENC0_IRQ
		EIE_INTIE |
#endif
		EIE_RXERIE );
	enc0_write_phy(PHIE, PHIE_PLNKIE | PHIE_PGEIE);
#ifdef ENC0_IRQ
	ENC0_IRQ_SETUP;
#endif

	/* Finally enable the packet reception */
	enc0_bfs_reg(ECON1, ECON1_RXEN);

	DEBUG_PUT(2, "ENC0 ready.");
}

static void enc0_soft_reset(void) {
	CS_ON;
	spi_putc(ENC0_OP_SRC);
	/* Wait for the device to become ready. Normally: Poll ESTAT.CLKRDY */
	/* But work around a silicon bug in revisions until B7: */
	_delay_ms(2);
	CS_OFF;
}

/*
 * enc0_select_bank:
 * Switches to the bank of register identified by address.
 * If the switch is not needed, it is not performed.
 */
static void enc0_select_bank(uint8_t address) {
	/* Calculate the bank number from an address: */
	uint8_t bank_no = (address & ENC0_REG_BANK_MASK) >> ENC0_REG_BANK_SHIFT;
	/* Is the switch needed? */
	if(((address & ENC0_REG_ADDR_MASK) < ENC0_REG_IN_ALL_BANKS) &&
		NE8(enc0_info.bank, bank_no)) {
		enc0_bfc_reg(ECON1, ECON1_BSEL1 | ECON1_BSEL0);
		enc0_bfs_reg(ECON1, bank_no);
		enc0_info.bank = bank_no;
	}
}

/*
 * enc0_read_reg:
 * Read the value of the supplied register (ETH, MAC, MII).
 * Bank switches are made if neccessary.
 */
static uint8_t enc0_read_reg(uint8_t address) {
	uint8_t value;
	/* Select Bank. */
	enc0_select_bank(address);

	CS_ON;
	/* Send the opcode and register address. */
	spi_putc(ENC0_OP_RCR | (address & ENC0_REG_ADDR_MASK));
	/* Receive the register value. */
	value = spi_putc(0);

	/* If we read a non-ETH register, the first value was only a dummy byte. */
	if((address & ENC0_REG_TYPE_MASK) != ENC0_REG_ETH)
		/* Read the real data. */
		value = spi_putc(0);

	CS_OFF;
	return value;
}

/*
 * enc0_wbfx_reg:
 * Write a register, or set/clear bits in an ETH register.
 * Since these operations only differ in their opcode, unify them.
 */
static void enc0_wbfx_reg(uint8_t opcode, uint8_t address, uint8_t value) {
	enc0_select_bank(address);

	CS_ON;
	spi_putc(opcode | (address & ENC0_REG_ADDR_MASK));
	spi_putc(value);
	CS_OFF;
}

/*
 * enc0_bfx_reg:
 * Write a register. Special case of enc0_wbfx, but it saves us
 * two bytes on each call (first argument).
 * Since these operations only differ in their opcode, unify them.
 */
static noinline void enc0_write_reg(uint8_t address, uint8_t value) {
	enc0_wbfx_reg(ENC0_OP_WCR, address, value);
}


/*
 * enc0_read_mem_byte:
 * Read one byte from the buffer memory.
 */
uint8_t enc0_read_mem_byte(void) {
	uint8_t value;
	CS_ON;
	spi_putc(ENC0_OP_RBM);
	value = spi_putc(0);
	CS_OFF;
	return value;
}

/*
 * enc0_read_mem_block:
 * More performant function to read a whole byte range.
 * The result is placed into the given buffer.
 */
void enc0_read_mem_block(void* buffer, uint16_t length) {
	CS_ON;
	spi_putc(ENC0_OP_RBM);
	do {
		*((uint8_t *)buffer++) = spi_putc(0);
	} while (--length);
	CS_OFF;
}

/*
 * enc0_write_mem_byte:
 * Read one byte from the buffer memory.
 */
void enc0_write_mem_byte(uint8_t value) {
	CS_ON;
	spi_putc(ENC0_OP_WBM);
	spi_putc(value);
	CS_OFF;
}

/*
 * enc0_start_write_mem_block:
 * Prequisite for calling enc0_write_mem_block().
 */
inline void enc0_start_write_mem_block(void) {
	CS_ON;
	spi_putc(ENC0_OP_WBM);
}

/*
 * enc0_stop_write_mem_block:
 * Finishes the memory writing. Must be called after enc0_write_mem_block.
 */
#define enc0_stop_write_mem_block() CS_OFF

/*
 * enc0_write_mem_block:
 * More performant function to write a whole byte range.
 * The data is taken from the given buffer.
 */
inline void enc0_write_mem_block(void* buffer, uint16_t length) {
	do {
		spi_putc(*(uint8_t*)buffer++);
	} while (--length);
}

/*
 * enc0_read_regpair:
 * Read a 16 bit value from a register pair.
 * This works as all register pairs are at consecutive positions.
 */
uint16_t enc0_read_regpair(uint8_t address) {
	uint16_t8 ret;
	ret.lo = enc0_read_reg(address);
	ret.hi = enc0_read_reg(++address);
	return ret.val;
}

/*
 * enc0_write_regpair:
 * Write a 16 bit value from a register pair.
 * This works as all register pairs are at consecutive positions.
 */
void enc0_write_regpair(uint8_t address, uint16_t value) {
	uint16_t8 val;
	val.val = value;
	enc0_write_reg(address, val.lo);
	enc0_write_reg(++address, val.hi);
}

/*
 * enc0_write_phy:
 * Write two bytes into a PHY register. A bit ugly to handle.
 */
static void enc0_write_phy(uint8_t address, uint16_t value) {
	/* Tell the MII the PHY address to write to. */
	enc0_write_reg(MIREGADR, address);

	/* Push out both data bytes. */
	enc0_write_regpair(MIWR, value);

	/* Finally tell the MII to write the bytes and wait until it is done. */
	while(enc0_read_reg(MISTAT) & MISTAT_BUSY);
}

/*
 * enc0_read_phy:
 * Read two bytes from a PHY register. Also ugly.
 */
static uint16_t enc0_read_phy(uint8_t address) {
	uint16_t data;
	/* Transfer the address to read from. */
	enc0_write_reg(MIREGADR, address);

	/* Tell the MII to read data from the PHY. */
	enc0_write_reg(MICMD, MICMD_MIIRD);

	/* Spin until data is available. */
	while(enc0_read_reg(MISTAT) & MISTAT_BUSY);

	/* Stop reading from the PHY */
	enc0_write_reg(MICMD, 0);

	data = enc0_read_regpair(MIRD);

	return data;
}

#ifdef ENC0_IRQ
ISR(ENC0_IRQ) {
	enc0_info.int_pending = 1;
}
#endif


void enc0_worker(void) {
#ifdef ENC0_IRQ
	if(!enc0_info.int_pending)
		return;
	enc0_info.int_pending = 0;
#endif

	/*
	 * Find out which interrupt sources fired.
	 */
	uint8_t int_sources = enc0_read_reg(EIR);

#ifdef ENC0_IRQ
	/* Disable interrupts until we finish. */
	enc0_bfc_reg(EIE, EIE_INTIE);
#endif

	/* Get count of pending packets */
	uint8_t pktcnt = enc0_read_reg(EPKTCNT);


	DEBUG_PRINT(3, "ENC0: Worker for EIR %x and %u packets\n", int_sources, pktcnt);
	if(!int_sources && !pktcnt) {
		DEBUG_PUT(2, "ENC0: Spurious interrupt.");
		goto end;
	}

	/* Did the link status change? */
	if(int_sources & EIR_LINKIF) {
		/* Clear interrupt flag. It may fire again. */
		enc0_read_phy(PHIR);

#if DEBUG_LEVEL
		if(enc0_read_phy(PHSTAT2) & PHSTAT2_LSTAT) {
			DEBUG_PUT(1, "ENC0: Link up");
		} else {
			DEBUG_PUT(1, "ENC0: Link down");
		}
#endif
	}

	if(int_sources & (EIR_TXIF | EIR_RXERIF | EIR_TXERIF)) {
#if DEBUG_LEVEL > 1
		uint8_t tx_stat = enc0_read_reg(ESTAT);
		if(tx_stat & ESTAT_LATECOL)
			DEBUG_PUT(2, "ENC0: Late collision.");
		if(tx_stat & ESTAT_TXABRT)
			DEBUG_PUT(2, "ENC0: Transmission aborted.");
		if(int_sources & EIR_RXERIF) {
			DEBUG_PUT(2, "ENC0: Reception failed.");
		}
		if(int_sources & EIR_TXERIF)
			DEBUG_PUT(2, "ENC0: Transmission failed 2.");
#endif
		/* Mark the TX circuit as idle if an transmission is finished: */
		if(int_sources & EIR_TXIF)
			enc0_info.tx_active = 0;

		/* Clear the interrupt flags. */
		enc0_bfc_reg(EIR, EIR_TXIF | EIR_RXERIF | EIR_TXERIF);
		enc0_bfc_reg(ESTAT, ESTAT_TXABRT | ESTAT_LATECOL);
	}


	if(int_sources & EIR_PKTIF || pktcnt > 0) {
		enc0_recv_packet();
	}

end:
#ifdef ENC0_IRQ
	/* Reenable interrupts. */
	enc0_bfs_reg(EIE, EIE_INTIE);
#endif
}


static void enc0_recv_packet(void) {

	/*
	 * Receive a maximum of 5 packets in a row to keep the latency for other
	 * tasks small.
	 */

	uint8_t loopcnt = 5;
	while(--loopcnt && (enc0_read_reg(EPKTCNT) > 0)) {
		enc0_rx_header rx_head;

		/*
		 * Set the pointer to the next packet. This is important if a
		 * packet is discarded.
		 */
		enc0_write_regpair(ERDPT, next_packet_loc);
		enc0_read_mem_block((void*)&next_packet_loc, sizeof(next_packet_loc));

		/* Get the RX header and check the size */
		enc0_read_mem_block((void*) &rx_head, sizeof(enc0_rx_header));
		/* Subtract the CRC length: */
		rx_head.packet_length -= 4;
		if(rx_head.packet_length < 14 ||
			rx_head.packet_length > MIN(UIP_MAX_ETH_FRAME_LENGTH, UIP_BUFSIZE) ||
			rx_head.unused || rx_head.crc_error || !rx_head.receive_ok
			) {
			DEBUG_PRINT(2, "ENC0: Discarded frame sized %u.\n", rx_head.packet_length);
			goto enc0_recv_packet_end;
		}

		/* Get the full packet as we want it */
		enc0_read_mem_block((void*) uip_buf, rx_head.packet_length);
		uip_len = rx_head.packet_length;

		struct uip_eth_hdr *eth_packet = (struct uip_eth_hdr*) uip_buf;
		switch(eth_packet->type) {
			case HTONS(UIP_ETHTYPE_ARP):
				uip_arp_arpin();
				if(uip_len > 0)
					enc0_tx_frame();
				break;

			case HTONS(UIP_ETHTYPE_IP):
				uip_arp_ipin();
				uip_input();
				if(uip_len > 0)
					uip_split_output();
				break;
			default:
				DEBUG_PUT(3, "ENC0: unknown Ethernet packet type");
		}

enc0_recv_packet_end:
		/*
		 * Clean up the current packet by advancing the receive read pointer
		 * and decreasing the packet counter.
		 * Implement a workaround for some obscure silicon bug:
		 * Enforce odd addresses in ERXRDPT.
		 */
		{
			uint16_t kludge_ptr = next_packet_loc - 1;
			if(kludge_ptr < ENC0_RXBUF_START || kludge_ptr > ENC0_RXBUF_END)
				kludge_ptr = ENC0_RXBUF_END;
			enc0_write_regpair(ERXRDPT, kludge_ptr);

			/* Decrement internal packet counter: */
			enc0_bfs_reg(ECON2, ECON2_PKTDEC);
		}
	}
}


/*
 * enc0_send_packet:
 * Add the ethernet header and then send an UIP packet.
 */
uint8_t enc0_send_packet() {
	uip_arp_out();
	return enc0_tx_frame();
}

#define tx_hw_active (enc0_read_reg(ECON1) & ECON1_TXRTS)
#define tx_dead (enc0_read_reg(EIR) & (EIR_TXERIF | EIR_TXIF))
#define tx_stamp enc0_last_tx_stamp = current_stamp
#define tx_delta \
	(timer_diff_timestamp16(&enc0_last_tx_stamp, &current_stamp))

/*
 * enc0_tx_frame:
 * Transmit an assembled frame that already has an Ethernet header.
 * returns 0: TX successfully initiated
 * otherwise: TX failed for some reason (busy?)
 */
uint8_t enc0_tx_frame(void) {
	timer_millistamp current_stamp;
	timer_get_timestamp(&current_stamp);

	DEBUG_PRINT(3, "ENC0: TX frame len %u\n", uip_len);
	/* Ignore too short packets, since they confuse the driver/hw: */
	if(uip_len < 14) return 1;


	/*
	 * For now, simply wait for any pending transmission to and and
	 * put any new packet at the start of the TX buffer.
	 */
	if(tx_hw_active) {
		if(tx_delta > 100)
			/* Sending took too long, reset the transmitter. */
			enc0_reset_tx();
		else {
			/* Spin a bit and hope that the transmit is finished then. */
			uint8_t timeout = 255;
			while((--timeout) && tx_hw_active);
			/* Oh, timeout: */
			if(!timeout) {
				DEBUG_PUT(1, "ENC0: TX timeout. Aborting.");
				return 1;
			}
		}
	}

	/* Packet sending starts at this address: */
	enc0_write_regpair(ETXST, ENC0_TXBUF_START);

	/* Packet length: */
	enc0_write_regpair(ETXND, ENC0_TXBUF_START + uip_len);

	/* The packet data is written to this address: */
	enc0_write_regpair(EWRPT, ENC0_TXBUF_START);

	/* Per packet control byte: */
	enc0_write_mem_byte(0);

	/* Packet copying: */
	enc0_start_write_mem_block();
	/* Ethernet header: */
	enc0_write_mem_block(&uip_buf[0], UIP_LLH_LEN);
	if(uip_len <= UIP_LLH_LEN + UIP_TCPIP_HLEN) {
		/* Shorter than TCP frame: */
		enc0_write_mem_block(&uip_buf[UIP_LLH_LEN], uip_len - UIP_LLH_LEN);
	} else {
		/* Longer than TCP frame, must copy TCP payload from uip_appdata: */
		enc0_write_mem_block(&uip_buf[UIP_LLH_LEN], UIP_TCPIP_HLEN);
		enc0_write_mem_block(uip_appdata, uip_len - UIP_TCPIP_HLEN - UIP_LLH_LEN);
	}
	enc0_stop_write_mem_block();

	/* Work around some other silicn bug: */
	enc0_reset_tx();

	/* Initiate transmit: */
	enc0_bfs_reg(ECON1, ECON1_TXRTS);

	/*
	 * Remember the current time in case we have to deal with a stuck
	 * TX engine. *cough*
	 */
	tx_stamp;
	enc0_info.tx_active = 1;
	enc0_info.tx_tries = 0;

	return 0;
}

/*
 * Microchip's engineers at their best.
 * Implement retransmissions in software for late collisions that
 * occur on the B5 and B7 revision in some cases.
 */
void enc0_retransmit(void) {
	if(enc0_info.tx_active && (enc0_hw_rev == 5 || enc0_hw_rev == 6)) {
		timer_millistamp current_stamp;
		timer_get_timestamp(&current_stamp);

		/* If the engine is busy for 10 ms, reset it. */
		if(tx_delta > 10 && tx_dead) {
			uint16_t read_ptr_save, tx_end;
			enc0_tx_header tx_head;

			if(enc0_info.tx_tries == 3) {
				/* Give up retrying after 3 tries. */
				DEBUG_PUT(1, "ENC: TX workaround giveup");
				enc0_reset_tx();
				return;
			}
			else
				DEBUG_PRINT(1, "ENC: TX workaround round %hu\n", enc0_info.tx_tries);

			enc0_info.tx_tries++;

			/* Cancel the current transmission as it became stuck. */
			enc0_bfc_reg(ECON1, ECON1_TXRST);

			/* Save the current read pointer. */
			read_ptr_save = enc0_read_regpair(ERDPTL);

			/* Get current transmit ptr, calculate TX status vector. */
			tx_end = enc0_read_regpair(ERDPTL);
			tx_end++;

			/* Address the status vector and get it: */
			enc0_write_regpair(ERDPT, tx_end);
			enc0_read_mem_block(&tx_head, sizeof(tx_head));

			if(tx_head.late_coll && (enc0_read_reg(EIR) & EIR_TXERIF)) {
				/* Guess: Reset the TX logic! */
				enc0_reset_tx();

				/* retransmit packet. */
				enc0_bfs_reg(ECON1, ECON1_TXRST);
				enc0_info.tx_active = 1;

				/* Implicitly wait for it to finish as we care in the next iteration. */
			}
		}
	}
}

/*
 * Thank god for crappy hardware. enc0_reset_tx is an important function
 * due to silicon bugs even in the latest revisions.
 */
void enc0_reset_tx(void) {
	/*
	 * Begin sending a packet and immediately abort the process.
	 * Clear the corresponding error and status interrupts.
	 */
	enc0_bfs_reg(ECON1, ECON1_TXRST);
	enc0_bfc_reg(ECON1, ECON1_TXRST);
	enc0_bfc_reg(EIR, EIR_TXERIF | EIR_TXIF);
	enc0_info.tx_active = 0;
}


/*
 * enc0_sputip:
 * Generate the dot-quatted representation of an IP address out of an UIP struct.
 */
char* enc0_sputip(char* dest, uip_ipaddr_t *ip) {
	uint8_t *ipptr = (uint8_t*) ip;
	uint8_t i = 4;
	do {
		dest = print_num16(dest, *ipptr++, 10);
		if(i != 1) *dest++ = '.';
	} while(--i);
	return dest;
}

/*
 * enc0_sscanip:
 * Generate an UIP struct out of an	dot-quatted IP address.
 * Returns 1 on success and 0 on failure.
 */
uint8_t enc0_sscanip(char* src, uip_ipaddr_t *ip) {
	uint32_t8 ipq;
	uint8_t i = 4;
	struct parse_ret16 ret;
	ret.next = src;
	do {
		ret = parse_num16(ret.next, 10);
		if(!ret.next || (i != 1 && *ret.next++ != '.')) return 0;
		ipq.b[4-i] = ret.val;
	} while(--i);
		*((uint32_t*)ip) = ipq.val;
		return 1;

}

/*
 * enc0_sputmac:
 * Generate the colon-separated representation of an MAC address.
 */
char* enc0_sputmac(char* dest, struct uip_eth_addr *eth) {
	uint8_t *macptr = (uint8_t*) eth;
	uint8_t i = 6;
	do {
		uint8_t byte = *macptr++;
		if(byte < 0x10) *dest++ = '0';
		dest = print_num16(dest, byte, 16);
		if(i != 1) *dest++ = ':';
	} while(--i);
	return dest;
}


/*
 * enc0_sscanmac:
 * Generate the UIP struct out of a MAC string.
 * Returns 1 on success and 0 on failure.
 */
uint8_t enc0_sscanmac(char* src, struct uip_eth_addr *eth) {
	struct uip_eth_addr tmp;
	uint8_t i = 6;
	struct parse_ret16 ret;
	ret.next = src;
	do {
		ret = parse_num16(ret.next, 16);
		if(!ret.next || (i != 1 && *ret.next++ != ':')) return 0;
		tmp.addr[6-i] = ret.val;
	} while(--i);
		*eth = tmp;
		return 1;
}

void enc0_update_mac(void) {
	uint8_t i = 0;
	uint8_t regno = MAADR1;
	do {
		uint16_t8 macpart;
		macpart.lo = uip_ethaddr.addr[i];
		macpart.hi = uip_ethaddr.addr[++i];
		enc0_write_regpair(regno, macpart.val);
		regno -= 2;
	} while (++i < 6);
}
