/*
 * Copyright (C) 2009-2010 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _NTPD_H
#define _NTPD_H

#include <inttypes.h>

extern void ntpd_init(void);
extern void ntpd_process_pkt(void);

/* This is from FreeBSD's /usr/src/contrib/ntp/include/ntp_fp.h: */

/*
 * NTP uses two fixed point formats.  The first (l_fp) is the "long"
 * format and is 64 bits long with the decimal between bits 31 and 32.
 * This is used for time stamps in the NTP packet header (in network
 * byte order) and for internal computations of offsets (in local host
 * byte order). We use the same structure for both signed and unsigned
 * values, which is a big hack but saves rewriting all the operators
 * twice. Just to confuse this, we also sometimes just carry the
 * fractional part in calculations, in both signed and unsigned forms.
 * Anyway, an l_fp looks like:
 *
 *    0       1         2       3
 *    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |             Integral Part           |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |             Fractional Part           |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 */
typedef struct {
  union {
    uint32_t Xl_ui;
    int32_t Xl_i;
  } Ul_i;
  union {
    uint32_t Xl_uf;
    int32_t Xl_f;
  } Ul_f;
} l_fp;

#define l_ui  Ul_i.Xl_ui    /* unsigned integral part */
#define l_i Ul_i.Xl_i   /* signed integral part */
#define l_uf  Ul_f.Xl_uf    /* unsigned fractional part */
#define l_f Ul_f.Xl_f   /* signed fractional part */


/* This is from FreeBSD's /usr/src/contrib/ntp/include/ntp.h: */

/*
 * NTP protocol parameters.  See section 3.2.6 of the specification.
 */
#define NTP_VERSION ((uint8_t)4) /* current version number */
#define NTP_OLDVERSION  ((uint8_t)1) /* oldest credible version */
#define NTP_PORT  123 /* included for non-unix machines */


/*
 * Values for peer mode and packet mode. Only the modes through
 * MODE_BROADCAST and MODE_BCLIENT appear in the transition
 * function. MODE_CONTROL and MODE_PRIVATE can appear in packets,
 * but those never survive to the transition function.
 * is a
 */
#define MODE_UNSPEC 0 /* unspecified (old version) */
#define MODE_ACTIVE 1 /* symmetric active mode */
#define MODE_PASSIVE  2 /* symmetric passive mode */
#define MODE_CLIENT 3 /* client mode */
#define MODE_SERVER 4 /* server mode */
#define MODE_BROADCAST  5 /* broadcast mode */

/*
 * Values for peer.leap, sys_leap
 */
#define LEAP_NOWARNING  0x0 /* normal, no leap second warning */
#define LEAP_ADDSECOND  0x1 /* last minute of day has 61 seconds */
#define LEAP_DELSECOND  0x2 /* last minute of day has 59 seconds */
#define LEAP_NOTINSYNC  0x3 /* overload, clock is free running */


#define PKT_MODE(li_vn_mode)  ((uint8_t)((li_vn_mode) & 0x7))
#define PKT_VERSION(li_vn_mode) ((uint8_t)(((li_vn_mode) >> 3) & 0x7))
#define PKT_LEAP(li_vn_mode)  ((uint8_t)(((li_vn_mode) >> 6) & 0x3))

#define PKT_LI_VN_MODE(li, vn, md) \
  ((uint8_t)((((li) << 6) & 0xc0) | (((vn) << 3) & 0x38) | ((md) & 0x7)))


/*
 * Dealing with stratum.  0 gets mapped to 16 incoming, and back to 0
 * on output.
 */
#define PKT_TO_STRATUM(s) ((uint8_t)(((s) == (STRATUM_PKT_UNSPEC)) ?\
        (STRATUM_UNSPEC) : (s)))

#define STRATUM_TO_PKT(s) ((uint8_t)(((s) == (STRATUM_UNSPEC)) ?\
        (STRATUM_PKT_UNSPEC) : (s)))

#define STRATUM_REFCLOCK ((uint8_t)0) /* default stratum */
/* A stratum of 0 in the packet is mapped to 16 internally */
#define STRATUM_PKT_UNSPEC ((uint8_t)0) /* unspecified in packet */
#define STRATUM_UNSPEC  ((uint8_t)16) /* unspecified */


/*
 * System event codes
 */
#define EVNT_UNSPEC 0 /* unspecified */
#define EVNT_SYSRESTART 1 /* system restart */
#define EVNT_SYSFAULT 2 /* wsystem or hardware fault */
#define EVNT_SYNCCHG  3 /* new leap or synch change */
#define EVNT_PEERSTCHG  4 /* new source or stratum */
#define EVNT_CLOCKRESET 5 /* clock reset */
#define EVNT_BADDATETIM 6 /* invalid time or date */
#define EVNT_CLOCKEXCPT 7 /* reference clock exception */


struct ntp_pkt {
  uint8_t  li_vn_mode; /* leap indicator, version and mode */
  uint8_t  stratum;  /* peer stratum */
  uint8_t  ppoll;    /* peer poll interval */
  signed char  precision;  /* peer clock precision */
  uint32_t  rootdelay;  /* distance to primary clock */
  uint32_t  rootdispersion; /* clock dispersion */
  uint32_t refid;    /* reference clock ID */
  l_fp  reftime;  /* time peer clock was last updated */
  l_fp  org;    /* originate time stamp */
  l_fp  rec;    /* receive time stamp */
  l_fp  xmt;    /* transmit time stamp */
	/* No extension included here */
};


#endif
