/*
 * Copyright (C) 2009-2010 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <avr/pgmspace.h>
#include <inttypes.h>
#include "avrutils.h"
#include "twi.h"
#include "saa1064.h"

/* How many times we try to send a message to a chip: */
#define SAA1064_TRIES 8

/* Addresses of the SAA1064 chips: */
#define SAA1064_0 0x70
/*#define SAA1064_1 0x72
 #define SAA1064_2 0x74
 #define SAA1064_3 0x76 */

#define CMD_CONTROL 0
#define CMD_DIGIT1  1

#define CTRL_MODE_DYNAMIC 0
#define CTRL_NOBLANK13 1
#define CTRL_NOBLANK24 2
#define CTRL_TEST 3
#define CTRL_CURRENT 4

/*
 * The 7 segment display is build this way:
 *                 a
 *           ,-----------,
 *           |           |
 *           |           |
 *         f |           | b
 *           |           |
 *           |     g     |
 *           +-----------+
 *           |           |
 *           |           |
 *         e |           | c
 *           |           |
 *           |           |
 *           '-----------'   DP
 *                 d
 */

/*
 * Represented characters: SP to _
     SP !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ
 * well-represented:
 *   SP  "    '    ,-  0123456789   = ? AbCdEF HI  L  OP  StU   y
 * guessable:
 *      ! #$%& ()*+  ./          :;< > @      g  jk      R    x  Z
 * unimplemented:
 *                                                  MN  Q    w
 */


#define SEG7_DP _BV(7)
uint8_t ascii_to_7seg(uint8_t ascii);
static const uint8_t saa1064_chars[] PROGMEM = {
#define a _BV(0) |
#define b _BV(1) |
#define c _BV(2) |
#define d _BV(3) |
#define e _BV(4) |
#define f _BV(5) |
#define g _BV(6) |
#define H SEG7_DP|
#define END 0,
	/* Start at ASCII 32 (space) */
 /* SP*/ END
 /* ! */ f b d  END
 /* " */ f b END
 /* # */ b c d e f g END
 /* $ */ a f g c d END
 /* % */ a b g f e END
 /* & */ e f a g b c END
 /* ' */ f END
 /* ( */ a f e d END
 /* ) */ a b c d END
 /* * */ a g d END
 /* + */ b c g END
 /* , */ e END
 /* - */ g END
 /* . */ d END
 /* / */ e g b END
 /* 0 */ a b c d e f END
 /* 1 */ b c END
 /* 2 */ a b g e d END
 /* 3 */ a b c d g END
 /* 4 */ f g b c END
 /* 5 */ a f g c d END
 /* 6 */ a f e d c g END
 /* 7 */ a b c END
 /* 8 */ a b c d e f g END
 /* 9 */ a b c d f g END
 /* : */ a d END
 /* ; */ c d END
 /* < */ g e d END
 /* = */ g d END
 /* > */ g d c END
 /* ? */ a b g e END
 /* @ */ g e c d END
 /* A */ e f a b g c END
 /* B */ f e d c g END
 /* C */ a f e d END
 /* D */ b c d e g END
 /* E */ a f e g d END
 /* F */ a f e g END
 /* G */ a c d e f END
 /* H */ b c g e f END
 /* I */ f e END
 /* J */ b c d END
 /* K */ e f g END
 /* L */ e f d END
 /* M */ g END
 /* N */ g END
 /* O */ c d e g END
 /* P */ a b g e f END
 /* Q */ g END
 /* R */ e g END
 /* S */ a f g c d END
 /* T */ f e g d END
 /* U */ b c d e f END
 /* V */ c d e END
 /* W */ g END
 /* X */  b c g e END
 /* Y */ f g b c d END
 /* Z */ a b g e d END
 /* [ */ d e f a END
 /* \ */ f g c END
 /* ] */ a b c d END
 /* ^ */ f a b END
 /* _ */ d END
#undef a
#undef b
#undef c
#undef d
#undef e
#undef f
#undef g
#undef H
};

struct send_arg{
	uint8_t chipno;
	uint8_t cmd;
	uint8_t data[2];
};

/*
 * saa1064_send_data:
 * Send data to the SAA1064 chip.
 * The destination chip number, command and additional data bytes
 * are given. In the case of a control command, only one 
 * data byte is sent.
 */
void saa1064_send_data(struct send_arg arg) {
	uint8_t i = SAA1064_TRIES;
	do {
		if(!twi_start(SAA1064_0 + 2*arg.chipno, TW_WRITE)) continue;
		/* Send control command: */
		TWDR = arg.cmd;
		if(twi_tx() == TW_MT_DATA_ACK) {
			/* Byte successfully sent, now transfer the control byte: */
			TWDR = arg.data[0];
			if(twi_tx() != TW_MT_DATA_ACK) goto send_stop;
			/* Only send the second byte if cmd is not the control command. */
			if(arg.cmd) {
				TWDR = arg.data[1];
				if(twi_tx() != TW_MT_DATA_ACK) goto send_stop;
			}
			i = 1; /* This ends the loop. */
		}
send_stop:
		twi_stop();
	} while (--i);
}

void saa1064_init(uint8_t chipno) {
	saa1064_set_brightness(chipno, 1); /* 3 mA current */
	uint16_t8 val; val.val = 0;
	saa1064_display(chipno, val); /* Clear display */
}

/*
 * saa1064_display:
 * Send the suppied value to the selected chip.
 * The value is an union of two bytes, which represent one segment each.
 * The supplied bytes are ASCII values and are transposed internally.
 */
void saa1064_display(uint8_t chipno, uint16_t8 val) {
	struct send_arg arg;
	arg.chipno = chipno;
	arg.cmd = CMD_DIGIT1;
	uint8_t i = 2;
	do {i--; arg.data[i] = ascii_to_7seg(val.b[i]);} while(i);
	saa1064_send_data(arg);
}

/*
 * saa1064_set_brightness:
 * Set the LED current of the addressed chip.
 * current bit values:
 * current:0:  3 mA
 * current:1:  6 mA
 * current:2: 12 mA
 * Its range is 3 to 21 mA, as you see.
 */
void saa1064_set_brightness(uint8_t chipno, uint8_t current) {
	struct send_arg arg;
	arg.chipno = chipno;
	arg.cmd = CMD_CONTROL;
	arg.data[0] = (0 << CTRL_MODE_DYNAMIC) | _BV(CTRL_NOBLANK13)
		| _BV(CTRL_NOBLANK24) | (current << CTRL_CURRENT);
	saa1064_send_data(arg);
}



uint8_t ascii_to_7seg(uint8_t ascii) {
	uint8_t dp_set = (ascii & _BV(7)) > 0;
	ascii &= ~_BV(7);
	if(ascii < ' ' || ascii > '_') ascii = ' ';
	uint8_t segment = pgm_read_byte(&saa1064_chars[ascii - ' ']);
	if(dp_set) segment |= SEG7_DP;
	return segment;
}
