/*
 * Copyright (C) 2009-2010 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <util/atomic.h>
#include <inttypes.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "config.h"
#include "timer.h"
#include "avrutils.h"

/* external: */
volatile date_time timer_localtime;
time_t timer_unix_timestamp, timer_uptime;
uint16_t timer_timestamp;
volatile struct timer_info timer_info;

enum day_state {IN_PAST = 0, TODAY, IN_FUTURE};

static const int  mon_lengths[MONSPERYEAR] PROGMEM =
  { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

static uint16_t skew_top;

static void timer_add_day(date_time*);
static enum day_state timer_months_last_sunday(uint8_t day, uint8_t dow);

/*
 * Determine if we are in the german DST range:
 * - from the last sunday in March, 02:00 CET
 * - till the last sunday in October, 02:00 CET
*/
static noinline uint8_t have_dst(const date_time *d, enum day_state last_sunday) {
	uint8_t is_dst = 0;
	uint8_t month = d->month;
	uint8_t hour = d->hour;
	if(month > 3 && month < 10) is_dst = 1;
	if (month ==  3 && (last_sunday == IN_PAST || (last_sunday == TODAY && hour >= 2)))
		is_dst = 1;
	if(month == 10 && (last_sunday == IN_FUTURE || (last_sunday == TODAY && hour <= 2)))
		is_dst = 1;
	return is_dst;
}

/*
 * ISR to drive the local clock:
 * Every decisecond, an interrupt is rised and handled.
 * This routine takes approx. < 3*64 clock cycles on mere second changes
 * and much more on hour and day changes.
 */
ISR(TIMER1_COMPA_vect) {
	add16_16(&timer_timestamp, 1);
	timer_info.dtick = 1;

	if(timer_info.skew_state) { /* skew is possible or applied */
		if(timer_info.skew_state == TIMER_SKEW_POSSIBLE) { /* modify the top value */
			OCR1A = skew_top;
			timer_info.skew_state = TIMER_SKEW_APPLIED;
		} else { /* The skew was applied */
			OCR1A = TIMER_LOAD_VALUE;
			timer_info.skew_state = TIMER_SKEW_IDLE;
		}
	}

	/* We get 10 interrupts per second, so count to 10 until a second is flipped */
	if(++timer_info.decisecond_counter == 10) {
		timer_info.decisecond_counter = 0;
		timer_info.tick = 1;

		++timer_uptime;

		/* If needed, count one further. This resembles a leap second. */
		uint8_t issue_leapsecond = timer_info.issue_leapsecond;
		uint8_t second_top = 60;
		if(issue_leapsecond) second_top = 61;

		if(++timer_localtime.second == second_top) {
			timer_localtime.second = 0;
			++timer_localtime.minute;
			if(issue_leapsecond) sub32_16((uint32_t*)&timer_unix_timestamp, 1);
			timer_info.issue_leapsecond = 0;
		}

		uint8_t hour = timer_localtime.hour;
		if(timer_localtime.minute == 60) {
			uint8_t new_dst_state;
  		enum day_state sunday_state = timer_months_last_sunday(timer_localtime.day, timer_localtime.dow);
			timer_localtime.minute = 0;
			hour++;

			timer_localtime.hour = hour; /* have_dst needs an up to date hour */
			/* Change the hour on DST change */
			new_dst_state = have_dst((date_time*) &timer_localtime, sunday_state);
			if(timer_localtime.dst != new_dst_state) {
				if(!timer_localtime.dst && new_dst_state)
					/* DST begins */
					hour++;
				else
					/* DST ends */
					hour--;

				timer_localtime.dst = new_dst_state;
			}
		}

		if(hour == 24) {
			hour = 0;
			/* Discard volatility (in ISR context): */
			timer_add_day((date_time*) &timer_localtime);
		}
		timer_localtime.hour = hour;

		add32_16((uint32_t*)&timer_unix_timestamp, 1);
	}
}

void timer_init(void) {
	ENTER_CRITICAL_SECTION
		/* Stop the timer */
		TCCR1B = 0;
		TCNT1 = 0;

		/* clear the timestamps */
		timer_timestamp = 0;
		timer_uptime = 0;
		timer_unix_timestamp=0;
		timer_unix2localtime(timer_unix_timestamp, (date_time*) &timer_localtime);
		/* Discard volatility attribute, as we are in a critical section: */
		bzero((struct timer_info*) &timer_info, sizeof(struct timer_info));

		/*
		 * Prescaler 64, internal clock source, compare match on TCNT1 == OCR1A
		 * Generate interrupt on match.
		*/
		OCR1A = TIMER_LOAD_VALUE;
		TCCR1A = 0;
		TCCR1B = (0<<WGM13) + (1<<WGM12) + (0<<CS12) + (1<<CS11) + (1<<CS10);
		TIMSK1 |= 1 << OCIE1A;
	LEAVE_CRITICAL_SECTION
}

/*
 * set_time:
 * Set the internal time to the new value.
 * Calculate the Unix timestamp from the given value and return it.
 */
void timer_set_time(const date_time *new_time) {
	ENTER_CRITICAL_SECTION
		timer_localtime = *new_time;
		timer_unix_timestamp = timer_localtime2unix(new_time);
		timer_info.valid = 1;
	LEAVE_CRITICAL_SECTION
}

/*
 * sync_ticks:
 * Reset the internal time counter to synchronize it to an external
 * reference. The supplied correction compensates delays introduced
 * by other parts of the system.
 */
void timer_sync_ticks(uint16_t correction) {
	ENTER_CRITICAL_SECTION
		TCNT1 = correction;
		timer_info.decisecond_counter = 0;

		/*
		 * Clear the corresponding interrupt flag, so that no interrupt
		 * is generated by a bad timing.
		 */
		TIFR1 |= (1 << OCF1A);
	LEAVE_CRITICAL_SECTION
}

/*
 * timer_sync_ticks:
 * Tell the ISR to modify the top value one time: */
void timer_skew_clock(int16_t diff) {
	/* Ensure that the argument is small enough: */
	if(diff < -((int16_t)TIMER_LOAD_VALUE)) diff = -((int16_t)TIMER_LOAD_VALUE);
	if(diff > (int16_t) TIMER_LOAD_VALUE) diff = (int16_t) TIMER_LOAD_VALUE;

	skew_top = TIMER_LOAD_VALUE + diff;
	timer_info.skew_state = TIMER_SKEW_POSSIBLE;
}

/*
 * timer_sync_ticks:
 * Atomically copy the current time.
 */
void timer_get_time(date_time *time) {
	ENTER_CRITICAL_SECTION
		*time = timer_localtime;
	LEAVE_CRITICAL_SECTION
}

/*
 * get_millis:
 * return a timer_timestamp with 1 millisecond per tick.
 */
void timer_get_timestamp(timer_millistamp *tm) {
	timer_millistamp temp;
	/*
	 * Ensure that the timer ISR has run by doing no operation
	 * some times. WARNING: the loop should count further if
	 * more interrupt handlers are used / fire often.
	 */
	temp.ms = TCNT1;
	if(temp.ms == 0 || temp.ms == TIMER_LOAD_VALUE) {
		uint8_t i = 10;
		do { asm volatile("nop\n"); } while(--i);
	}
	/*
	 * Now, the ISR ran and timer_info.decisecond_counter and timer_unix_timestamp
	 * are in sync with TCNT1.
	 */
	ENTER_CRITICAL_SECTION
		temp.ms = TCNT1;
		temp.ds = timer_timestamp;
	LEAVE_CRITICAL_SECTION
	if(temp.ms > TIMER_LOAD_VALUE) temp.ms = TIMER_LOAD_VALUE;
	*tm = temp;
}

/*
 * timer_diff_timestamp:
 * Return the passed time from t1 to t2 in milliseconds.
 */
uint32_t timer_diff_timestamp(const timer_millistamp *t1, const timer_millistamp *t2) {
	uint32_t diff;
	stamp_t dsdiff = 0;
	/* Overflow handled implicitly: */
	dsdiff = t2->ds - t1->ds;

	/* Handle millisecond overflows specially: */
	diff = t2->ms;
	if(t2->ms < t1->ms)
		dsdiff -= 1, diff += TIMER_SCALE_VALUE;
	diff -= t1->ms;
	diff *= 100;
	diff /= TIMER_SCALE_VALUE;
	diff += dsdiff * 100;

	return diff;
}

/*
 * timer_diff_timestamp16:
 * Light-weight variant of timer_diff_timestamp.
 * Returns ~0 even if the diff is > (uint16_t)(~0).
 */
uint16_t timer_diff_timestamp16(const timer_millistamp *t1, const timer_millistamp *t2) {
	uint32_t val = timer_diff_timestamp(t1, t2);
	if(val > 0xFFFF)
		return 0xFFFF;
	else
		return (uint16_t) val;
}

/*
 * timer_get_pstamp:
 * Return a precise Unix timestamp.
 * The seconds part is tha normal unix timestamp,
 * but the fractions part is a binary fraction.
 * Meaning, fraction == 2^32 - 1 is equivalent to 0.99... seconds.
 */
void timer_get_pstamp(timer_pstamp *stampP) {
	uint8_t decis;
	/*
	 * Ensure that the timer ISR has run by doing no operation
	 * some times. WARNING: the loop should count further if
	 * more interrupt handlers are used / fire often.
	 */
	uint16_t tcnt_temp = TCNT1;
	if(tcnt_temp == 0 || tcnt_temp == TIMER_LOAD_VALUE) {
		uint8_t i = 10;
		do { asm volatile("nop\n"); } while(--i);
	}
	/*
	 * Now, the ISR ran and timer_info.decisecond_counter and timer_unix_timestamp
	 * are in sync with TCNT1.
	 */
	uint32_t fraction;
	ENTER_CRITICAL_SECTION
		fraction  = TCNT1;
		decis = timer_info.decisecond_counter;
		stampP->seconds = timer_unix_timestamp;
	LEAVE_CRITICAL_SECTION
	if(fraction > TIMER_LOAD_VALUE) fraction = TIMER_LOAD_VALUE;
#define MAX_FRACTION 4294967296UL
#define FRACTION_STEP (uint32_t)(MAX_FRACTION/10)
#define MICRO_SHIFT 16;
#define MICRO_MULT (uint32_t) 65536UL
#define MICRO_STEP (uint32_t)(MAX_FRACTION/MICRO_MULT/10)
	/* Microsecond to 100 ms section: */
	fraction <<= MICRO_SHIFT;
	fraction /= TIMER_SCALE_VALUE;
	fraction *= MICRO_STEP;
	/* Decisecond section: */
	stampP->fraction = fraction + ((uint32_t) decis) * FRACTION_STEP;
}


static noinline void timer_add_day(date_time *d) {
	uint8_t leap_year = timer_is_leap_year(d->year);
	uint8_t d_month = d->month - 1;
	uint8_t days_of_month = pgm_read_byte(&mon_lengths[d_month]);
	if(leap_year && d_month == 1) days_of_month++;

	uint8_t dow = d->dow + 1;
	d->dow = dow % DAYSPERWEEK;
	d->day++;

	if( d->day > days_of_month) {
		d->day = 1;
		++d->month;
	}
	if(d->month > 12) {
		d->month = 1;
		++d->year;
	}
}


static void timer_unix2utc(const time_t unix_time, date_time *d) {
	time_t epoch_secs = unix_time; /* needed to declare the argument as const */
	uint16_t day, month, year;

	/*
	 * The time conversion to UTC is straight forward, as the Unix timestamp
	 * starts at 00:00:00 UTC. So split out the day part as division remainders.
	*/
	d->second  = epoch_secs % SECSPERMIN;
	epoch_secs = epoch_secs / SECSPERMIN;

	d->minute  = epoch_secs % MINSPERHOUR;
	epoch_secs = epoch_secs / MINSPERHOUR;

	d->hour    = epoch_secs % HOURSPERDAY;
	epoch_secs = epoch_secs / HOURSPERDAY;

	/*
	 * day now contains the number of days since the Unix Epoch.
	 * Its maximal value is 2^32 / 60 / 60 / 24 < 49711, it fits into an uint16_t.
	 * This should speed up the following operations.
	 * Correct day to let ist start with 1 instead of 0.
	 */
	day = epoch_secs + 1;

	/*
	 * The day of week can simply be obtained if we know that the 1st January
	 * 1970 was a Thursday.
	 */
	d->dow = (EPOCH_DOW + day - 1) % DAYSPERWEEK;

	/*
	 * Simply loop through the year from 1970 until there are less days left
	 * than a year has. This is not elegant, but simple.
	 */
	year = EPOCH_YEAR;

	for(year = EPOCH_YEAR; day >= DAYSPERNYEAR; year++) {
		if(timer_is_leap_year(year)) {
			/* The loop condition only guarantees days >= 365, not >= 366. */
			if(day >= DAYSPERLYEAR)
				/* We have enough days left to end this year. */
				day -= DAYSPERLYEAR;
			else
				/*
				 * If we have 365 days left in a leap year, it is not over
				 * and we are finished in this loop.
				 */
				break;
		} else {
			day -= DAYSPERNYEAR;
		}
	}

	/* Convert the day of year to a month:day pair. */
	month = 0;
	uint8_t leap_year = timer_is_leap_year(year);
	do {
		uint8_t this_months_days =
			pgm_read_byte(&mon_lengths[month]);
		if(leap_year && month == 1) this_months_days++;

		/*
		 * Try to subtract this month's days from the day counter.
		 * If not enough days are left, the month is determined.
		 */
		if(day < this_months_days) break;

		day -= this_months_days;
		month++;
	} while(1);

	d->day   = day;
	d->month = month + 1;
	d->year  = year;
}


time_t timer_localtime2unix(const date_time *d) {
	time_t unix_t = 0;
	uint8_t month;
	uint16_t year;
	/*
	 * Inverse the process of timer_unix2utc.
	 */

	/* Add the seconds for the years 1970 ... d->year to unix_t: */
	for(year = EPOCH_YEAR; year < d->year; year++) {
		if(timer_is_leap_year(year))
			unix_t += (time_t) SECSPERDAY * DAYSPERLYEAR;
		else
			unix_t += (time_t) SECSPERDAY * DAYSPERNYEAR;
	}

	/*
	 * Add the seconds for the months 1 ... d->month of d->year
	 * Internal, months count from 0, not 1. Correct it initially.
	 */
	uint8_t leap_year = timer_is_leap_year(year);
	for(month = d->month - 1; month > 0; month--) {
		uint8_t this_months_days =
			pgm_read_byte(&mon_lengths[month - 1]);
		if(leap_year && month == 2) this_months_days++;
		/* Leap years are handled by the mon_lengths array. */
		unix_t += this_months_days * SECSPERDAY;
	}

	uint8_t d_second = d->second;
	uint8_t d_minute = d->minute;
	uint8_t d_hour = d->hour;
	uint8_t d_day = d->day;
	unix_t += ((time_t)(((d_day - 1) * HOURSPERDAY + d_hour)) * MINSPERHOUR
		+ d_minute) * SECSPERMIN;
	unix_t += d_second;

	/*
	 * Since we know if DST is active, correct the time zone and DST offset
	 * at last. Hope to use a 16 bit multiplication as a 12 hour offset
	 * fits into uint16_t, but not int16_t.
	 */

	int16_t offset = TIMER_TZ;
	if(offset < 0)
		unix_t += ((uint16_t) offset) * SECSPERMIN;
	else
		unix_t -= ((uint16_t) offset) * SECSPERMIN;

	if(d->dst)
		unix_t -= SECSPERHOUR;

	return unix_t;
}

/*
 * timer_months_last_sunday:
 * Return where the last sunday of a month with 31 days is
 * relative to the given day and day of week.
 */
static noinline enum day_state timer_months_last_sunday(uint8_t day, uint8_t dow)
{
  uint8_t last_satureday = day;
	/* Determine the satureday before "day": */
  if (last_satureday > dow) {
    last_satureday -= dow; /* Sunday before day */
		last_satureday--; /* Satureday */
	}
	/*
	 * else: There is no previous satureday in this month.
	 * => last_satureday remains very small.
	 * So, this can't be the last sunday in the month.
	 */


  if (last_satureday >= 31 - 7)
		/* Beyond the 24th, there can't be another sunday in this month. */
		if (dow == 0)
			/* Whoo, today /is/ sunday. It must be the last in this month. */
			return TODAY;
		else
	    return IN_PAST;
  else /* There must be another sunday in this month since more than one week is left: */
	  return IN_FUTURE;
}

void timer_unix2localtime(const time_t t_in, date_time *d) {
	time_t t = t_in;

	/* Add the time zone offset: */
	t += SECSPERMIN * ((time_t) TIMER_TZ);

	/* Check if we are in the DST range */
	timer_unix2utc(t, d);
  enum day_state sunday_state = timer_months_last_sunday(d->day, d->dow);
  if (have_dst(d, sunday_state)) {
		/* As we have DST, add the DST offset: */
    t += SECSPERHOUR;
		d->dst = 1;
  } else
		d->dst = 0;

	/* Generate the date/time again with DST and time zone correction: */
	timer_unix2utc(t, d);
}


uint8_t timer_is_leap_year(uint16_t year) {
	return (!(year % 4) && ((year % 100) || !(year % 400)));
}

uint16_t noinline timer_ctr_delay2ticks(uint16_t delay) {
	return (uint16_t) timer_ctr_us2ticks(delay);
}

uint16_t noinline timer_ctr_ticks2delay(uint16_t ticks) {
	return (uint16_t) timer_ctr_ticks2us(ticks);
}

void timer_add_leapsecond(void) {
	timer_info.issue_leapsecond = 1;
}
