/*
 * Copyright (C) 2009-2010 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _DS_TIMER_H
#define _DS_TIMER_H

#include <inttypes.h>


#define TIMER_PRESCALER_VALUE 64

/* Calculate the value at which a compare match interrupt is generated: */
#define TIMER_LOAD_VALUE (F_CPU / TIMER_PRESCALER_VALUE / 10 - 1)
/* Divide TCNT1 through this value to obtain a relative amount: */
#define TIMER_SCALE_VALUE (F_CPU / TIMER_PRESCALER_VALUE / 10)

#if (TIMER_LOAD_VALUE+1) * 10 * TIMER_PRESCALER_VALUE != F_CPU
#error "F_CPU is unsuited for the timer operation (too small, odd?)"
#endif

#define EPOCH_YEAR	1970
#define EPOCH_DOW   4 /* Thursday */
#define SECSPERMIN  60
#define MINSPERHOUR 60
#define HOURSPERDAY 24
#define DAYSPERWEEK 7
#define DAYSPERNYEAR  365
#define DAYSPERLYEAR  366
#define SECSPERHOUR (SECSPERMIN * MINSPERHOUR)
#define SECSPERDAY  ((uint32_t) SECSPERHOUR * HOURSPERDAY)
#define MONSPERYEAR 12

typedef struct {
	uint8_t second;
	uint8_t minute;
	uint8_t hour:5;
	uint8_t dst:1; /* daylight savings enabled? */
	uint8_t day; /* day of month (1-32) */
	uint8_t month:4;
	uint8_t dow:4; /* 0 is sunday */
	uint16_t year; /* year of century */
} date_time;

/*
 * time_t is unsigned. Consequences:
 * - There is no 2038 problem here.
 * - Dates before 1970 cannot be handled.
 */
typedef uint32_t time_t;
typedef uint16_t stamp_t;

#define TIMER_SKEW_IDLE 0
#define TIMER_SKEW_POSSIBLE 1
#define TIMER_SKEW_APPLIED 2

struct timer_info {
	uint8_t decisecond_counter:4;
	uint8_t skew_state:2;
	uint8_t valid:1;
	uint8_t tick:1;
	uint8_t dtick:1;
	uint8_t issue_leapsecond:1;
};

typedef struct {
	stamp_t ds;
	uint16_t ms;
} timer_millistamp;

/* return a precise timestamp: */
typedef struct {
	uint32_t seconds;
	uint32_t fraction;
} timer_pstamp;

/* decoded time in CET / CEST */
extern volatile date_time timer_localtime;

/*
 * seconds since the Unix epoch (here: unsigned)
 * This field is volatile, but you must enter a critical section before accessing it anyway.
 */
extern time_t timer_unix_timestamp;

/*
 * seconds since the system was booted.
 * This field is volatile, but you must enter a critical section before accessing it anyway.
 */
extern time_t timer_uptime;

/*
 * A simple counter with 100 milliseconds per tick
 * This field is volatile, but you must enter a critical section before accessing it anyway.
 */
extern stamp_t timer_timestamp;

/* A flag if the second has changed (clear it yourself) */
extern volatile struct timer_info timer_info;

/*
 * timer_init:
 * Initialize internal time counting
 */
void timer_init(void);

/*
 * timer_set_time:
 * Set the internal time to the new value
 * is_cest: set if the time is CEST, not CET
 */
void timer_set_time(const date_time *new_time);

/*
 * timer_sync_ticks:
 * Reset the internal time counter to synchronize it to an external clock
 */
void timer_sync_ticks(const uint16_t correction);


/*
 * timer_skew_clock:
 * Add the given number of ticks one time to the timer's counter top value.
 */
void timer_skew_clock(const int16_t diff);


/*
 * timer_get_time:
 * Atomically fill in the current time.
 */
void timer_get_time(date_time *time);

/*
 * timer_get_millis:
 * Writes a timestamp with 1 millisecond resolution to
 * the specified location.
 */
void timer_get_timestamp(timer_millistamp *target);

/*
 * timer_diff_timestamp:
 * Return the number of milliseconds between t2 and t1
 */
uint32_t timer_diff_timestamp(const timer_millistamp *t1, const timer_millistamp *t2);

/*
 * timer_diff_timestamp16:
 * Return the number of milliseconds between t2 and t1.
 * This is the lighter-weight vairiant of timer_diff_timestamp.
 * If the difference is larger than the maximal return value,
 * the maximal return value is returned.
 */
uint16_t timer_diff_timestamp16(const timer_millistamp *t1, const timer_millistamp *t2);

/*
 * timer_get_pstamp:
 * Return a precise Unix timestamp with binary fraction.
 */
void timer_get_pstamp(timer_pstamp*);

/*
 * timer_add_leapsecond:
 * Add a leap second at the end of this minute.
 */
void timer_add_leapsecond(void);

/* 
 * timer_unix2localtime:
 * convert a unix timestamp t to a date/time structure d
 * offset_minutes: GMT/UTC offset of the local time
 */
void timer_unix2localtime(const time_t t, date_time *d);

/*
 * timer_localtime2unix:
 * convert a local time (including DST) to a unix timestamp
 * offset_minutes: GMT/UTC offset of the local time
 */
time_t timer_localtime2unix(const date_time *d);

/*
 * is_leap_year:
 * Returns 1 if the year is a leap year, otherwise 0.
 */
uint8_t timer_is_leap_year(uint16_t year);

/*
 * Macro to retrieve a counter value:
 */
#define timer_get_ctr(var) \
	do {ENTER_CRITICAL_SECTION \
		var = TCNT1; \
	LEAVE_CRITICAL_SECTION \
	} while(0)

/*
 * Macros for the conversion from a counter delay in microseconds to
 * counter-understandable delays and back 
 */
#define timer_ctr_us2ticks(delay) \
	(((uint32_t)(delay)) * TIMER_SCALE_VALUE / 100000)

#define timer_ctr_ticks2us(ticks) \
	(((uint32_t)(ticks)) * 100000 / TIMER_SCALE_VALUE)


/* Functions to do the above (for dynamic arguments): */
uint16_t timer_ctr_delay2ticks(uint16_t delay);
uint16_t timer_ctr_ticks2delay(uint16_t ticks);

/*
 * Macro do advance a counter by the given amount of ticks
 * and handle potential rollovers.
 * XXX: This works only if ticks + var does not overflow uint16_t!
 */
#define timer_ctr_add_ticks(var, ticks) \
	do {var += ticks; var %= TIMER_SCALE_VALUE;} while(0)

#endif
